/*
 * Generate a VGA signal using a PIC32 microcontroller.
 *
 * Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mips.h"
#include "pic32.h"

#define HFREQ_LIMIT	627	/* 20MHz cycles */
#define VSYNC_START	1	/* horizontal lines, front porch end */
#define VISIBLE_START	4	/* horizontal lines */
#define VBP_START	516	/* horizontal lines, back porch start */
#define VBP_END		531	/* horizontal lines, back porch end */

/* Disable JTAG functionality on pins. */

.section .devcfg0, "a"
.word 0xfffffffb	/* DEVCFG0<2> = JTAGEN = 0 */

/*
Set the oscillator to be the FRC oscillator with PLL, with peripheral clock
divided by 1, HS oscillator mode selected (for PLL), and FRCDIV+PLL selected.
With a system clock of 20MHz, the peripheral clock is therefore 20MHz.

The watchdog timer (FWDTEN) is also disabled.
*/

.section .devcfg1, "a"
.word 0xff7fcff9	/* DEVCFG1<23> = FWDTEN = 0; DEVCFG1<13:12> = FPBDIV<2:0> = 0; DEVCFG1<2:0> = FNOSC<2:0> = 001 */

/*
Set the FRC oscillator PLL function with an input division of 4, an output
division of 2, a multiplication of 20, yielding a multiplication of 2.5.
This should take the FRC at 8MHz and produce a system clock of 20MHz.
*/

.section .devcfg2, "a"
.word 0xfff9ffdb	/* DEVCFG2<18:16> = FPLLODIV<2:0> = 1; DEVCFG2<6:4> = FPLLMUL<2:0> = 101; DEVCFG2<2:0> = FPLLIDIV<2:0> = 011 */

.text
.globl _start

_start:
	/*
	Configure RAM.
	See: http://microchipdeveloper.com/32bit:mx-arch-exceptions-processor-initialization
	*/

	la $v0, BMXCON
	li $v1, (1 << 6)	/* BMXCON<6> = BMXWSDRM = 0 */
	sw $v1, CLR($v0)

	/* Enable caching. */

	li $v0, CONFIG_CM_CACHABLE_NONCOHERENT
	mtc0 $v0, CP0_CONFIG
	nop

	/* Get the RAM size. */

        la $v0, BMXDRMSZ
	lw $v0, 0($v0)

	/* Initialise the stack pointer. */

	lui $v1, 0x8000		/* 0x80000000 */
	addu $sp, $v0, $v1	/* sp = 0x80000000 + RAM size */

	/* Initialise the globals pointer. */

	lui $gp, %hi(_GLOBAL_OFFSET_TABLE_)
	ori $gp, $gp, %lo(_GLOBAL_OFFSET_TABLE_)

	/* Set pins for output and set outputs low. */

	jal init_pins
	nop

	/* Initialise the status register. */

	jal init_interrupts
	nop

	/* Initialise timer. */

	jal init_timer1
	nop

	/* Enable interrupts and loop. */

	jal enable_interrupts
	nop

	jal handle_error_level
	nop

	li $a1, 20000000	/* counter = 20000000 */
	li $a2, 6374		/* test timer scaling */
	li $s0, 0		/* line counter */
	la $s1, vfp_active	/* current event */
loop:
	addiu $a1, $a1, -1      /* counter -= 1 */
	bnez $a1, loop          /* until counter == 0 */
	nop

	li $a1, 20000000	/* counter = 20000000 */

        la $v0, PORTA
	la $v1, (1 << 2)	/* PORTA<2> = RA2 */
	sw $v1, INV($v0)

_next:
	j loop
	nop



init_pins:
	/* DEVCFG0<2> needs setting to 0 before the program is run. */

        la $v0, CFGCON
	li $v1, (1 << 3)	/* CFGCON<3> = JTAGEN = 0 */
	sw $v1, CLR($v0)

	la $v0, TRISB
	li $v1, (7 << 7) | 15	/* TRISB<9:7> = RB9, RB8, RB7 = 0 */
	/* sw $v1, CLR($v0) */

        la $v0, PORTB
	li $v1, (7 << 7) | 15	/* PORTB<9:7> = RB9, RB8, RB7 = 0 */
	/* sw $v1, CLR($v0) */

init_leds:
        la $v0, TRISA
	li $v1, (1 << 2)	/* TRISA<2> = RA2 = 0 */
	sw $v1, CLR($v0)

	la $v0, TRISB
	li $v1, (3 << 10)	/* TRISB<11:10> = RB11, RB10 = 0 */
	sw $v1, CLR($v0)

clear_leds:
        la $v0, PORTA
	li $v1, (1 << 2)	/* PORTA<2> = RA2 = 0 */
	sw $v1, CLR($v0)

        la $v0, PORTB
	li $v1, (3 << 10)	/* PORTB<11:10> = RB11, RB10 = 0 */
	sw $v1, CLR($v0)

	jr $ra
	nop



/* Utilities. */

handle_error_level:
	mfc0 $t3, CP0_STATUS
	li $t4, ~(STATUS_ERL | STATUS_EXL)
	and $t3, $t3, $t4
	mtc0 $t3, CP0_STATUS
	jr $ra
	nop

enable_interrupts:
	mfc0 $t3, CP0_STATUS
	nop
	ori $t3, $t3, STATUS_IRQ | STATUS_IE
	mtc0 $t3, CP0_STATUS
	jr $ra
	nop

init_interrupts:
	mfc0 $t3, CP0_STATUS
	li $t4, ~STATUS_BEV
	and $t3, $t3, $t4
	mtc0 $t3, CP0_STATUS	/* CP0_STATUS &= ~STATUS_BEV (use non-bootloader vectors) */

	la $t3, _start
	mtc0 $t3, CP0_EBASE	/* EBASE = _start */

	li $t3, CAUSE_IV	/* IV = 1 (use EBASE+0x200 for interrupts) */
	mtc0 $t3, CP0_CAUSE

	jr $ra
	nop



/* Interrupt servicing. */

.org 0x200

interrupt_handler:

	/* Check for a timer interrupt condition. */

	la $v0, IFS0
	lw $v1, 0($v0)
	and $v1, $v1, (1 << 4)	/* T1IF */
	beqz $v1, irq_exit
	nop

	/* Test timer scaling. */

	addiu $a2, $a2, -1
	bnez $a2, irq_clear
	nop
	li $a2, 6374

	/* Increment the line counter. */

	addiu $s0, $s0, 1

	/* Jump to the event handler. */

	jalr $s1
	nop

irq_clear:

	/* Clear the timer interrupt condition. */

	la $v0, IFS0
	li $v1, (1 << 4)	/* IFS0<4> = T1IF = 0 */
	sw $v1, CLR($v0)

irq_exit:
	eret
	nop



/* Initialisation routines. */

init_timer1:

	/* Initialise Timer1 interrupt. */

	la $v0, T1CON
	sw $zero, 0($v0)	/* T1CON = 0 */
	nop

	la $v0, TMR1
	sw $zero, 0($v0)	/* TMR1 = 0 */
	la $v0, PR1
	li $v1, HFREQ_LIMIT
	sw $v1, 0($v0)		/* PR1 = HFREQ_LIMIT */

	/* Initialise Timer1 interrupt. */

	la $v0, IFS0
	li $v1, (1 << 4)
	sw $v1, CLR($v0)	/* IFS0CLR: T1IF = 0 */
	la $v0, IPC1
	li $v1, (7 << 2)
	sw $v1, SET($v0)	/* IPC1SET: T1IP = 7 */
	la $v0, IPC1
	li $v1, 3
	sw $v1, SET($v0)	/* IPC1SET: T1IS = 3 */
	la $v0, IEC0
	li $v1, (1 << 4)
	sw $v1, SET($v0)	/* IEC0SET: T1IE = 1 */

	/* Start timer. */

	la $v0, T1CON
	li $v1, (1 << 15)	/* ON = 1 */
	sw $v1, SET($v0)	/* T1CONSET: ON = 1 */

	jr $ra
	nop



/* Event routines. */

/* Start of, and within, the vertical front porch. */

vfp_start:
	/* Clear vsync. */

        la $v0, PORTB
	la $v1, (1 << 10)	/* PORTB<10> = RB10 */
	sw $v1, CLR($v0)

	/* Enter the active region. */

	la $s1, vfp_active

vfp_active:
	/* Test for vsync. */

	subu $v0, $s0, VSYNC_START
	bnez $v0, _vfp_active_ret
	nop

	/* Start the vsync region. */

	la $s1, vsync_start

_vfp_active_ret:
	jr $ra
	nop



/* Start of, and within, vsync. */

vsync_start:
	/* Set vsync. */

        la $v0, PORTB
	la $v1, (1 << 10)	/* PORTB<10> = RB10 */
	sw $v1, SET($v0)

	/* Enter the active region. */

	la $s1, vsync_active

vsync_active:
	/* Test for visible region. */

	subu $v0, $s0, VISIBLE_START
	bnez $v0, _vsync_active_ret
	nop

	/* Start the visible region. */

	la $s1, visible_start

_vsync_active_ret:
	jr $ra
	nop



/* Start of, and within, the visible period. */

visible_start:
	/* Clear vsync. */

        la $v0, PORTB
	la $v1, (1 << 10)	/* PORTB<10> = RB10 */
	sw $v1, CLR($v0)

	/* Enter the active region. */

	la $s1, visible_active

visible_active:
	/* NOTE: Perform line activities. */

        la $v0, PORTB
	la $v1, (1 << 11)	/* PORTB<11> = RB11 */
	sw $v1, INV($v0)

	/* Test for back porch. */

	subu $v0, $s0, VBP_START
	bnez $v0, _visible_active_ret
	nop

	/* Start the back porch. */

	la $s1, vbp_active

_visible_active_ret:
	jr $ra
	nop



/* Within the vertical back porch. */

vbp_active:
	/* Test for front porch. */

	subu $v0, $s0, VBP_END
	bnez $v0, _vbp_active_ret
	nop

	/* Start the front porch. */

	li $s0, 0
	la $s1, vfp_start

_vbp_active_ret:
	jr $ra
	nop
